(******************************************************************************

  Modeling and Verifying Properties of Security Protocols in Wireless Networks

  Module: distance/AuthenticatedRanging.thy (Isabelle/HOL 2009)

  Copyright (C) 2007-9 Benedikt Schmidt <benedikt.schmidt@inf.ethz.ch>
  License: BSD (see COPYRIGHT)

******************************************************************************)

header {* Security Analysis of an Authenticated Ranging Protocol *}

theory AuthenticatedRanging
imports SystemCoffset SystemOrigination MessageTheoryFree
begin

locale INITSTATE_SIG_NN = INITSTATE_PKSIG + INITSTATE_NONONCE

text {* here we define the initstate for the protocol, then interpret it as \_SIG\_NN *}

definition
  initStateAR :: "agent \<Rightarrow> msg set" where
  "initStateAR A == Key`(insert (priSK A) (pubSK`UNIV))"

interpretation INITSTATE_SIG_NN Crypt Nonce MPair Hash Number parts subterms DM initStateAR Key
  apply (unfold_locales)
  apply (auto simp add: initStateAR_def)
done

subsection {* Direct and Parameterized Definition and their Equality *}

subsubsection {* Step Functions for Authenticated Ranging *}

definition
  ar1 :: "msg step"
 where
 "ar1 tr A t =
    (UN NA. {ev. ev = (Nonce (Honest A) NA,SendEv 0 []) \<and>
                 Nonce (Honest A) NA \<notin> usedI tr})"

definition
  ar2 :: "msg step"
 where
 "ar2 tr B tsend =
  (UN NA C trec.
          {ev.
           ev = (Crypt (priSK (Honest B))
                       \<lbrace>Nonce C NA,Real (tsend - trec)\<rbrace>,SendEv 0 []) \<and>

           (trec, Recv (Rx (Honest B) 0) (Nonce C NA)) \<in> set tr})"

definition
  ar3 :: "msg step"
where
 "ar3 tr A t = 
    (UN NA B tsend delta trec.
        {ev. ev = (\<lbrace>Agent B, Real ((trec - tsend - delta) * vc / 2)\<rbrace>, ClaimEv) \<and>
         (trec, Recv (Rx (Honest A) 0)
                         (Crypt (priSK B)
                               \<lbrace>Nonce (Honest A) NA, Real delta\<rbrace>)) \<in> set tr \<and>
         (tsend, Send (Tr (Honest A)) (Nonce (Honest A) NA) []) \<in> set tr})"

definition
  ar_proto :: "msg proto" where
  "ar_proto = {ar1,ar2,ar3}"

lemmas ar_defs = ar_proto_def ar1_def ar2_def ar3_def

text {* The initial state contains no nonces and signatures, additionally, it includes
        the necessary public and private keys and keys are never sent over the network. *}

locale PROTOCOL_AR =
  PROTOCOL_PKSIG_NOKEYS+PROTOCOL_NONONCE+INITSTATE_SIG_NN

interpretation PROTOCOL_AR Crypt Nonce MPair Hash Number parts subterms DM initStateAR Key ar_proto
  apply (unfold_locales)
  apply (auto simp add: ar_defs messagesProto_def messagesProtoTrHonest_def
              split: event.split split_if dest: parts.fst_set)
done

text {* Agents only look at their own views and all messages are derivable. *}

interpretation PROTOCOL_EXECUTABLE Crypt Nonce MPair Hash Number parts subterms DM initStateAR  ar_proto sys Key
  apply (unfold_locales)
  apply (auto simp add: ar_defs DM_analz_synth initStateAR_def
                        messagesProto_def messagesProtoTrHonest_def
              split: event.split split_if)

  apply (drule view_elem_ex, erule exE, drule Recv_imp_knows_A)
  apply (auto intro!: synth.Crypt)
  apply (auto simp add: nonce_view_fresh [simplified ar_proto_def]
                        nonce_view_used [simplified ar_proto_def])
  apply (auto simp add: recv_a_view_a_r send_a_view_a_r)
done

text {* Agent behaviour does not change with constant clock errors. *}

interpretation PROTOCOL_DELTAONLY Crypt Nonce MPair Hash Number parts subterms DM initStateAR Key ar_proto
  apply unfold_locales
  apply (auto simp add: ar_defs in_timetrans)
  apply force prefer 2
  apply (rule_tac x="trec + coffset A" in exI, force)
  apply (intro exI conjI)
  apply auto
  apply (rule exI)
  apply (rule_tac x="tsend + coffset A" in exI, rule exI,
         rule_tac x="trec + coffset A" in exI)
  apply auto
done

interpretation PROTOCOL_DELTA_EXEC Crypt Nonce MPair Hash Number parts subterms DM initStateAR Key ar_proto sys
by unfold_locales

subsubsection {* Direct Definition for Authenticated Ranging *}

inductive_set
  ranging :: "(msg trace) set"

 where
   Nil [intro] : "[] \<in> ranging"

 | Fake: 
   "\<lbrakk> tr \<in> ranging; t \<ge> maxtime tr;
      X \<in> DM (Intruder I) (knowsI (Intruder I) tr)\<rbrakk>
   \<Longrightarrow> (t, Send (Tx (Intruder I) j) X [])  # tr \<in> ranging"

 | Con : 
   "\<lbrakk> tr \<in> ranging; trecv \<ge> maxtime tr;
      (tsend, Send (Tx A i) m L) \<in> set tr;
      cdistM (Tx A i) (Rx B j) = Some tab;
      trecv \<ge> tsend + tab \<rbrakk>
    \<Longrightarrow> (trecv, Recv (Rx B j) m) # tr \<in> ranging"

 -- {* Protocol rules for authenticated ranging, correspond to step functions *}
 | AR1:  
   "\<lbrakk> tr \<in> ranging; t >=  maxtime tr ;
      \<not> (usedI tr (Nonce (Honest A) NA)) \<rbrakk> 
        (* cannot use set notation here: infer_arities: bad term (Isabelle Bug?) *)
   \<Longrightarrow> (t, Send (Tx (Honest A) 0) (Nonce (Honest A) NA) []) # tr \<in> ranging"

 | AR2:  
   "\<lbrakk> tr \<in> ranging;  tsend \<ge>  maxtime tr;
      (trec, Recv (Rx (Honest B) 0) (Nonce A NA)) \<in> set tr \<rbrakk>
   \<Longrightarrow> (tsend,
        Send (Tx (Honest B) 0) (Crypt (priSK (Honest B))
                                      \<lbrace>Nonce A NA, Real (tsend - trec)\<rbrace>) [])
      # tr \<in> ranging"

 | AR3:  
   "\<lbrakk> tr \<in> ranging; t \<ge> maxtime tr;
      (trec, Recv (Rx (Honest A) 0) (Crypt (priSK B)
                                           \<lbrace>Nonce (Honest A) NA, Real delta\<rbrace>))
      \<in> set tr;
      (tsend, Send (Tr (Honest A)) (Nonce (Honest A) NA) []) \<in> set tr\<rbrakk>
  \<Longrightarrow> (t, Claim (Honest A) \<lbrace>Agent B, Real ((trec - tsend - delta)*vc/2)\<rbrace>) # tr
      \<in> ranging"

text {* obtain a simpler induction rule for protocol since it is executable and deltaonly *}

lemmas proto_induct =
  sys.induct [simplified derivable_removable remove_occursAt timetrans_removable]

subsubsection {* Equality for direct and Parameterized Definition *}

text {* We now show that both inductive definitions define the
        same set of traces. *}

lemma abstr_equal: "ranging = sys"
proof auto
  fix tr
  assume r: "tr \<in> sys"
  show "tr \<in> ranging" using r
  proof (induct tr rule: proto_induct)
    case 1 with prems show ?case by auto
  next
    case 2 with prems show ?case by (auto intro: ranging.Nil)
  next
    case 4 with prems show ?case by (auto intro: ranging.Con)
  next
    case 3 with prems show ?case by (auto intro: ranging.Fake)
  next
    case (5 tr t step m pEv A)
    thus ?case
      apply (auto simp add: ar_defs)
      apply (auto intro!: ranging.AR1 ranging.AR2 ranging.AR3 [simplified])
      apply (simp add: mem_def)
      done
  qed
next
  fix tr
  assume r: "tr \<in> ranging"
  show "tr \<in> sys" using r
  proof(induct tr rule: ranging.induct)
    case Nil
    with prems show ?case by auto
  next
    case (Fake tr ts X I j)
    with prems show ?case by (auto intro: sys.Fake)
  next
    case (Con tr tsend C i X D j dCD trecv)
    with prems show ?case by (auto intro: sys.Con)
  next
    case (AR1 tr ts C NA)
    with prems have "(ts,createEv C (SendEv 0 []) (Nonce (Honest C) NA)) # tr \<in> sys"
      apply -
      apply (rule_tac step=ar1 in sys_Proto_exec)
      apply (auto simp add: ar_defs)
      apply (simp add: mem_def)
      done
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (AR2 tr tsend trecv B A NA)
    with prems have
      "(tsend, 
        createEv B
                 (SendEv 0 [])
                 (Crypt (priSK (Honest B))
                    \<lbrace>Nonce A NA, Real (tsend - trecv)\<rbrace>)) # tr \<in> sys"
      apply - apply (rule_tac step=ar2 in sys_Proto)
      apply (auto simp add: ar_defs)
      done
    thus ?case by (auto simp add: createEv.psimps)
  next
    case (AR3 tr t trecv A B NA delta tsend)
    with prems have
      "(t,
        createEv A ClaimEv \<lbrace>Agent B, Real ((trecv - tsend - delta)*vc/2)\<rbrace>) # tr \<in> sys"
      apply - apply (rule_tac step=ar3 in sys_Proto)
      apply (auto simp add: ar_defs)
      done
    thus ?case by (auto simp add: createEv.psimps)
  qed
qed

lemmas [simp,intro] = abstr_equal [THEN sym]

subsection {* A "possibility property": The protocol completes succesfully in certain scenarios*}

lemma ranging_possibility:
 assumes distab: "cdistM (Tr (Honest A)) (Rec (Honest B)) \<noteq> None"
 assumes distba: "cdistM (Tr (Honest B)) (Rec (Honest A)) \<noteq> None"
 shows "\<exists>ta dt. \<exists>tr \<in> ranging. (ta,Claim (Honest A) \<lbrace>Agent (Honest B), Real dt\<rbrace>)
        \<in> set tr"
proof -
  let ?dab = "cdist (Tr (Honest A)) (Rec (Honest B))" and
      ?dba = "cdist (Tr (Honest B)) (Rec (Honest A))" and
      ?fa = "Honest A" and
      ?fb = "Honest B" and
      ?na = "Nonce (Honest A) NA"
  let ?msg2 = "Crypt (priSK ?fb) \<lbrace>?na, Real (?dab-?dab)\<rbrace>" and
      ?done = "Claim ?fa \<lbrace>Agent ?fb, Real (((?dab+?dba) - 0 - 0)*vc/2)\<rbrace>"
  from distab have dabnoneg: "?dab \<ge> 0"
    by (auto simp add: cdist_def intro: cdistnoneg_some)
  from distba have dbanoneg: "?dba \<ge> 0"
    by (auto intro: cdistnoneg_some simp add: cdist_def)
  from dabnoneg dbanoneg have snoneg: "?dab+?dba \<ge> 0" by auto
  have "(0,Send (Tr ?fa) ?na [])#[] \<in> ranging" (is "?tr1 \<in> ranging")
    apply - apply (rule ranging.AR1)
    apply auto
    apply (auto simp add: usedI_def)
    by (simp add: mem_def [THEN Pure.symmetric, where x1="Nonce (Honest A) NA"])
  with distab dabnoneg have "(0+?dab,Recv (Rec ?fb) ?na)#?tr1 \<in> ranging" (is "?tr2 \<in> ranging")
     apply -
     apply (rule_tac tab="cdist (Tr (Honest A)) (Rec (Honest B))" in ranging.Con)
     by (auto simp add: cdist_def)
  with dabnoneg have "(?dab,Send (Tr ?fb) ?msg2 [])#?tr2 \<in> ranging"
    (is "?tr3 \<in> ranging") apply -
    by (rule_tac tr="?tr2" in ranging.AR2, auto)
  hence "(?dab+?dba,Recv (Rec ?fa) ?msg2)#?tr3 \<in> ranging"
    (is "?tr4 \<in> ranging")
    using snoneg distba dbanoneg apply - apply (rule ranging.Con)
    by (auto simp add: cdist_def)
  hence "(?dab+?dba, ?done)#?tr4 \<in> ranging" (is "?tr5 \<in> ranging")
    apply (rule_tac tr="?tr4" in ranging.AR3, auto)
    apply (rule dbanoneg) apply (rule snoneg) done
  -- {* This is the witness trace: @{term [display] ?tr5} *}
  thus ?thesis
    apply (rule_tac x="cdist (Tr (Honest A)) (Rec (Honest B))
                       +cdist (Tr (Honest B)) (Rec (Honest A))" in exI)
    apply (rule_tac x="(cdist (Tr (Honest A)) (Rec (Honest B)) +
                         cdist (Tr (Honest B)) (Rec (Honest A)) - 0 - 0) * vc / 2" in exI)
    apply (intro bexI) by auto
qed

subsection {* Invariants concerning signatures and nonces by honest agents*}

lemma recv_before_send_sig:
  assumes rang: "tr \<in> ranging" and
          A: "(tb, Send (Tx (Honest B) i) mb L) \<in> set tr" and
          B: "Crypt (priSK (Honest Ba)) \<lbrace>Nonce A NA, Real tdelta\<rbrace>
              \<in> subterms {mb}"
  shows   "(tb - tdelta, Recv (Rx (Honest B) 0) (Nonce A NA)) \<in> set tr"
  using rang A B
  apply (induct tr)
  apply (auto)
done

lemma nonce_send_honest:
  assumes send: "(ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) L) \<in> set tr" and
          rang: "tr \<in> ranging"
  shows   "Nonce (Honest A) NA
           \<notin> usedI (beforeEvent (ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) L) tr)"
  using rang send
proof (induct rule: ranging.induct) 
  case Nil thus ?case by auto
next
  case (Con tr tc C mc D tab)
  hence "(ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) L) \<in> set tr" by auto
  with Con.hyps prems show ?case by (auto)
next
  case (Fake tr mintr I tsend)
  hence "(ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) L) \<in> set tr" by auto
  with Fake.hyps prems show ?case by (auto)
next
  case (AR2 tr t trecv D NB)
  hence "(ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) L) \<in> set tr" by auto
  with AR2.hyps prems show ?case by (auto)
next
  case (AR3 tr t trecv D E NB tdelta tsend)
  hence "(ta, Send (Tx (Honest A) i) (Nonce (Honest A) NA) L) \<in> set tr" by auto
  with AR3.hyps prems show ?case by (auto)
next
  case (AR1 tr NB tsend D)
  with AR1.hyps prems show ?case apply auto apply (auto simp add: mem_def)  done
qed

subsection {* Security Proof for Authenticated Ranging *}

lemma authranging_secure:
  assumes
  rang: "tr \<in> ranging" and aneqb: "A\<noteq>B" and
  believe: "(tdone, Claim (Honest A) \<lbrace>Agent (Honest B), Real d\<rbrace>) \<in> set tr"
  shows   "d \<ge> pdist (Honest A) (Honest B)"
  using prems
proof (induct tr arbitrary: A B trec t rule: ranging.induct)
  -- {* the only nontrivial case since it adds Claim events *}
  case (AR3 tr t trecv D E NA tdelta tsend)
  let ?x = "(t, Claim (Honest D) \<lbrace>Agent E, Real ((trecv - tsend - tdelta)*vc/2)\<rbrace>)"
  and ?ev = "((tdone, Claim (Honest A) \<lbrace>Agent (Honest B), Real d\<rbrace>))"
  show ?case proof cases
    -- {* the added event is the Claim event from the premise,
          the other case follows trivially from the IH *}
    assume "?x = ?ev"
    hence "A=D" and "E=Honest B" and "t=tdone" and tdelta:"d=(trecv - tsend - tdelta)*vc/2"
      by auto
    let ?sigmsg =  "(Crypt (priSK (Honest B)) \<lbrace>Nonce (Honest A) NA, Real tdelta\<rbrace>)"
    from `A=D` `E=Honest B` prems have
      ononce: "(tsend, Send (Tr (Honest A)) (Nonce (Honest A) NA) []) \<in> set tr" and
      sig: "(trecv, Recv (Rec (Honest A)) ?sigmsg) \<in> set tr"
      by auto
    with `tr \<in> ranging` obtain C tcsend i L
      where sigintrC: "(tcsend, Send (Tx C i) ?sigmsg L) \<in> set tr" and
                      "trecv >= cdist (Tx C i) (Rec (Honest A))" and
                      "cdistM (Tx C i) (Rec (Honest A)) \<noteq> None" and
            q: "tcsend \<le> trecv - cdist (Tx C i) (Rec (Honest A))"
      apply - apply (drule send_before_recv [simplified])
      by auto
    let ?tsigsendC = "trecv - cdist (Tx C i) (Rec (Honest A))"
    have "?sigmsg \<in> subterms {?sigmsg}" and  "(Honest B) \<notin> Intruder`UNIV" by auto
    show ?thesis
    proof cases
      -- {* If the signature comes directly from honest user B, then we know
            that there is a Recv NA at B and a Send NA event at A with
	    the needed time differences. *}
      assume "C=Honest B"
      with prems sigintrC have
        sigrecvB: "(tcsend - tdelta, Recv (Rx (Honest B) 0) (Nonce (Honest A) NA)) \<in> set tr"
          apply - apply (rule recv_before_send_sig) apply (auto) done
      from `?sigmsg \<in> subterms {?sigmsg}` have "Nonce (Honest A) NA \<in> subterms {?sigmsg}" by auto
      moreover have "(Honest A) \<notin> Intruder`UNIV" by auto
      moreover
        have non: "Nonce (Honest A) NA
                   \<notin> usedI (beforeEvent (tsend, Send (Tr (Honest A))  (Nonce (Honest A) NA)[]) tr)"
        using prems apply - apply (rule nonce_send_honest) by auto
      moreover have "Nonce (Honest A) NA \<in> subterms {Nonce (Honest A) NA}" by (auto intro: subterms.Inj)
      ultimately have tbta: "tcsend - tdelta - tsend  >= cdistl (Honest A) (Honest B)"
        using `tr \<in> ranging` `A\<noteq>B`
              `(tsend, Send (Tr (Honest A)) (Nonce (Honest A) NA) []) \<in> set tr` sigrecvB
        apply (erule_tac tr=tr in fresh_nonce_earliest_recv [simplified])
        by (force simp add: usedI_def)+
      -- {* needed inequalities *}
      from `cdistM (Tx C i) (Rec (Honest A)) \<noteq> None` noflt_some `C=Honest B`
        have "cdistl (Honest B) (Honest A) \<le> cdist (Tx C i) (Rec (Honest A))"
	  by (auto simp add: cdist_def)
      with q have "cdistl (Honest B) (Honest A) \<le> trecv - tcsend" by auto
      with tbta  `C=Honest B`
        have "cdistl (Honest A) (Honest B) + cdistl (Honest B) (Honest A) 
              \<le> trecv - tsend - tdelta " by auto
      with cdistl_symm
        have "cdistl (Honest A) (Honest B) + cdistl (Honest A) (Honest B)
              \<le> trecv - tsend - tdelta " by auto
      hence "2 *cdistl (Honest A) (Honest B) \<le> trecv - tsend - tdelta " by auto
      hence "cdistl (Honest A) (Honest B) \<le> (trecv - tsend - tdelta) /2" by auto
      hence "cdistl (Honest A) (Honest B) * vc \<le> (trecv - tsend - tdelta)/2 * vc" apply -
	thm mult_right_mono
	apply (rule mult_right_mono)
	by (auto, insert vc_pos, auto)
      hence "pdist (Honest A) (Honest B) \<le> (trecv - tsend - tdelta) * vc/2"
        apply (auto simp add: cdistl_def split: split_if_asm)
        by (insert vc_pos, auto) moreover
      from `?x=?ev` have "(trecv - tsend - tdelta) * vc/2 = d" by auto
      ultimately show ?thesis by auto
    next
      -- {* If the signature comes from an Intruder, then he must have received
            the Signature from B and B must have received the nonce sent by A *}
      assume "C\<noteq>Honest B"
      with sigintrC `tr \<in> ranging`
      have "\<exists>tb mb i Lb. tcsend - tb >= cdistl (Honest B) C \<and>
        (tb, Send (Tx (Honest B) i) mb Lb) \<in> set tr \<and>
        ?sigmsg \<in> subterms {mb}" apply - apply (rule sig_earliest_send)
        by auto
      then obtain tb mb k Lb where
	"tcsend - tb >= cdistl (Honest B) C" and
	sigsendB: "(tb, Send (Tx (Honest B) k) mb Lb) \<in> set tr" and
	sigmb: "?sigmsg \<in> subterms {mb}"
	by auto
      with q have tb: " ?tsigsendC -tb >= cdistl (Honest B) C" by auto
      with prems have
        sigrecvB: "(tb - tdelta, Recv (Rx (Honest B) 0) (Nonce (Honest A) NA)) \<in> set tr"
	apply - apply (rule recv_before_send_sig) apply (auto)
        done
      from sigmb have "Nonce (Honest A) NA \<in> subterms {?sigmsg}" and "{?sigmsg} \<subseteq> subterms {mb}"
        by auto
      hence "Nonce (Honest A) NA \<in> subterms {mb}"
        apply - apply (rule_tac G="{?sigmsg}" in subterms.trans)
	by auto
      moreover
      have "(Honest A) \<notin> Intruder`UNIV" by auto moreover
      have non: "Nonce (Honest A) NA
        \<notin> usedI (beforeEvent (tsend, Send (Tr (Honest A))  (Nonce (Honest A) NA) []) tr)"
	using prems apply - apply (rule nonce_send_honest) by auto
      have "Nonce (Honest A) NA \<in> subterms {Nonce (Honest A) NA}" by (auto intro: subterms.Inj)
      ultimately have tbta: "(tb - tdelta) - tsend  >= cdistl (Honest A) (Honest B)"
	using `tr \<in> ranging` `A\<noteq>B` and non and
          `(tsend, Send (Tr (Honest A)) (Nonce (Honest A) NA) []) \<in> set tr` and sigrecvB
           apply - apply (rule_tac tr=tr and j=0 and mb="Nonce (Honest A) NA"
                          in fresh_nonce_earliest_recv)
           apply (simp add: abstr_equal [THEN sym]) by (force simp add: usedI_def)+
      have "cdistl (Honest B) (Honest A) \<le>
        cdistl (Honest B) C + cdistl C (Honest A)" by (rule cdistl_triangle)
      also from `cdistM (Tx C i) (Rec (Honest A)) \<noteq> None` noflt_some have
	"\<dots> \<le> cdist (Tx C i) (Rec (Honest A)) + cdistl (Honest B) C"
	by (auto simp add: cdist_def)
      also from tb have "\<dots> \<le> trecv - tb" by auto
      finally have "cdistl (Honest B) (Honest A) \<le> trecv - tb" by auto
      with tbta have "cdistl (Honest A) (Honest B) + cdistl (Honest B) (Honest A)
        \<le> trecv - tsend - tdelta " by auto moreover
      with cdistl_symm have "cdistl (Honest A) (Honest B) + cdistl (Honest A) (Honest B)
        \<le> trecv - tsend - tdelta " by auto
      hence "2 *cdistl (Honest A) (Honest B) \<le> trecv - tsend - tdelta " by auto
      hence "cdistl (Honest A) (Honest B) \<le> (trecv - tsend - tdelta) /2" by auto
      hence "cdistl (Honest A) (Honest B) * vc \<le> (trecv - tsend - tdelta) * 1/2 * vc"
       apply -
       apply (rule mult_right_mono)
       by (auto, insert vc_pos, auto)
     hence "pdist (Honest A) (Honest B) \<le> (trecv - tsend - tdelta) * vc/2"
        apply (auto simp add: cdistl_def split: split_if_asm)
        by (insert vc_pos, auto) moreover
      from `?x=?ev` have "(trecv - tsend - tdelta) * vc/2 = d" by auto
      ultimately show ?thesis by auto
    qed
  next
    -- {* trivial case: just use induction hypothesis *}
    assume "?x \<noteq> ?ev"
    with `?ev \<in> set (?x#tr)` have "?ev \<in> set tr" by auto
    with prems show ?thesis by auto
  qed
next
  case Nil thus ?case by auto
next
  case (Con tr tc C mc D tab)
  hence "((tdone, Claim (Honest A) \<lbrace>Agent (Honest B), Real d\<rbrace>)) \<in> set tr" by auto
  with Con.hyps prems show ?case by (auto)
next
  case (Fake tr mintr I tsend)
  hence "((tdone, Claim (Honest A) \<lbrace>Agent (Honest B), Real d\<rbrace>)) \<in> set tr" by auto
  with Fake.hyps prems show ?case by (auto)
next
  case (AR2 tr trecv D NB tsend)
  hence "((tdone, Claim (Honest A) \<lbrace>Agent (Honest B), Real d\<rbrace>)) \<in> set tr" by auto
  with AR2.hyps prems show ?case by (auto)
next
  case (AR1 tr NB tsend D)
  hence "((tdone, Claim (Honest A) \<lbrace>Agent (Honest B), Real d\<rbrace>)) \<in> set tr" by auto
  with AR1.hyps prems show ?case by (auto)
qed

end
