(******************************************************************************

  Modeling and Verifying Properties of Security Protocols in Wireless Networks

  Module: distance/Distance.thy (Isabelle/HOL 2009)

  Copyright (C) 2007-2009 Benedikt Schmidt <benedikt.schmidt@inf.ethz.ch>
  License: BSD (see COPYRIGHT)

******************************************************************************)

header {* Physical Distance and Communication Distance *}

theory Distance imports Event CauchySchwarz begin

text {* some physical constants of our model: the speed of light a
nd sound,
        dimension of the space (2 or 3, but we can prove everything for $n$) *}

consts
  vu :: real
  vc :: real
  sdim :: nat

specification (vc)
  vc_pos: "vc > 0"
by (rule_tac x=1 in exI, arith)

specification (vu)
  vu_pos: "vu > 0"
by (rule_tac x=1 in exI, arith)

text {* $loc$ returns the location of an agent as a real vector of dimension $sdim$ *}

consts
  loc :: "agent \<Rightarrow> vector"

specification (loc)
 loc_dim: "vlen (loc A) = sdim"
by (rule_tac x="\<lambda>A. (\<lambda>n. 0, sdim)" in exI, auto simp add: vlen_def)

text {* we need vector subtraction for deriving the pseudometric from the real-norm *}

definition
  minusv :: "vector \<Rightarrow> vector \<Rightarrow> vector"                  ("_ -: _" 100) where
  "minusv v w = (\<lambda>n. v\<^sub>n  - w\<^sub>n,sdim)"

text {* we need vector addition in some proofs *}
definition
  plusv :: "vector \<Rightarrow> vector \<Rightarrow> vector"                  ("_ +: _" 100) where
  "plusv v w = (\<lambda>n. v\<^sub>n  + w\<^sub>n ,sdim)"

text {* relative physical distance between two agents, derived from  location function*}
definition
  pdist :: "[agent, agent] \<Rightarrow> real"
 where
  "pdist A B = \<parallel> loc A -: loc B \<parallel>"

text {* Line-of-Sight communication distance with speed of light *}
definition
  cdistl :: "[agent, agent] \<Rightarrow> real"
 where
 "cdistl A B = pdist A B / vc"

text {* pdist is a pseudometric *}
lemma pdist_noneg:
  "pdist A B \<ge> 0"
by (unfold pdist_def, rule norm_pos)

lemma square_minus_comm:
 "((a::real) - b)^2 = (b - a)^2"
proof -
  have "(a - b)^2 = (a - b)*(a -b)" by (simp add: real_sq)
  also have "\<dots> = a*a - a*b -b*a +b*b" by (simp add: real_sub_mult_distrib_ex)
  also have "\<dots> = b* b -b*a - a*b + a*a" by simp
  also have "\<dots> = (b - a)*(b -a)" by (simp add: real_sub_mult_distrib_ex) 
  also have "\<dots> = (b - a)^2" by (simp add: real_sq)
  finally show ?thesis by assumption
qed

lemma pdist_symm:
  "pdist A B = pdist B A"
by (unfold pdist_def norm_def minusv_def vlen_def ith_def, simp add: square_minus_comm) 

(* ToDo: use locales *)
definition
  zerov :: "vector" where
  "zerov = (\<lambda>n. 0, sdim)"

lemma vequal:
 "\<lbrakk> vlen v = vlen w; fst v = fst w \<rbrakk> \<Longrightarrow> v = w"
by (case_tac v, simp add: vlen_def)

lemma zerov_zero_plus:
  "loc A +: zerov = loc A" 
apply (simp add: plusv_def zerov_def ith_def vlen_def)
apply (rule vequal)
apply (simp add: loc_dim)
apply (simp add: vlen_def)
by (simp)

lemma minus_equal_zero:
  "loc A -: loc A = zerov"
by (auto simp add: minusv_def zerov_def ith_def vlen_def)

lemma pdist_equal_zero:" pdist A A = 0"
apply (simp add: pdist_def)
apply (simp add: minusv_def)
apply (simp add: norm_def)
apply (simp add: ith_def)
done

lemma minusv_comm:
  "loc A +: loc B = loc B +: loc A"
by (simp add: plusv_def ith_def, rule ext, simp)

lemma v_assoc1:
  "loc A +: (loc B -: loc B) = (loc A -: loc B) +: loc B"
apply (simp add: minus_equal_zero)
apply (simp only: zerov_zero_plus)
apply (simp add: plusv_def minusv_def)
apply (simp add: ith_def)
apply (rule vequal)
apply (simp add: loc_dim)
apply (simp add: vlen_def)
by simp

lemma v_assoc2:
  "((loc A -: loc B) +: loc B) -: loc C = (loc A -: loc B) +: (loc B -: loc C)"
apply  (auto simp add: plusv_def minusv_def)
by (rule ext, auto simp add: ith_def)

lemma norm_triangle:
  assumes vdim: "vlen v = sdim" and wdim: "vlen w = sdim"
  shows "\<parallel>v +: w\<parallel> \<le> \<parallel>v\<parallel> + \<parallel>w\<parallel>"
  using vdim wdim
proof -
 have "\<parallel>v +: w\<parallel>^2 \<le> (\<parallel>v\<parallel> + \<parallel>w\<parallel>)^2" proof -
   have "\<parallel>v +: w\<parallel>^2 = (\<Sum>k\<in>{1..sdim}. (v\<^sub>k + w\<^sub>k)^2)" using norm_pos
     apply (simp add: norm_def plusv_def ith_def vlen_def)
     by (auto simp add: norm_def ith_def vlen_def) 
   also have "\<dots> = (\<Sum>k\<in>{1..sdim}. (v\<^sub>k + w\<^sub>k)*(v\<^sub>k + w\<^sub>k))"
     by (auto simp add: real_sq)
   also have "\<dots> = (\<Sum>k\<in>{1..sdim}. v\<^sub>k*v\<^sub>k + v\<^sub>k* w\<^sub>k
                                  +w\<^sub>k*v\<^sub>k + w\<^sub>k*w\<^sub>k)"
     by (auto simp add: real_add_mult_distrib_ex)
   also have "\<dots> = (\<Sum>k\<in>{1..sdim}. v\<^sub>k*v\<^sub>k)
                 + (\<Sum>k\<in>{1..sdim}. v\<^sub>k* w\<^sub>k)
                 + (\<Sum>k\<in>{1..sdim}. w\<^sub>k*v\<^sub>k)
                 + (\<Sum>k\<in>{1..sdim}. w\<^sub>k*w\<^sub>k)"
     by (simp only: setsum_addf)
   also have "\<dots> = \<parallel>v\<parallel>^2 + (w\<cdot>v) + (v\<cdot>w) + \<parallel>w\<parallel>^2" using vdim wdim apply -
     apply (insert norm_pos, auto simp add: norm_def real_sqrt_pow2 dot_def)
     apply (auto simp add: real_sq)
     by (case_tac v, case_tac w, auto simp add: real_sqrt_pow2 vlen_def)
   also have "\<dots> = \<parallel>v\<parallel>^2 + (w\<cdot>v) + (w\<cdot>v) + \<parallel>w\<parallel>^2" using vdim wdim apply -
     by (auto simp add: dot_def RealDef.real_mult_commute)
   also have "\<dots> = \<parallel>v\<parallel>^2 + 2*(w\<cdot>v) + \<parallel>w\<parallel>^2" by auto
   also have "\<dots> \<le>  \<parallel>v\<parallel>^2 + 2*\<bar>w\<cdot>v\<bar> + \<parallel>w\<parallel>^2" by auto
   also have "\<dots> \<le>  \<parallel>v\<parallel>^2 + 2*\<parallel>v\<parallel>*\<parallel>w\<parallel> + \<parallel>w\<parallel>^2"
     apply (auto) apply (subst RealDef.real_mult_commute)
     apply (rule CauchySchwarzReal)
     by (insert vdim wdim, auto)
   also have "\<dots> = \<parallel>v\<parallel>*\<parallel>v\<parallel> + \<parallel>v\<parallel>*\<parallel>w\<parallel> + \<parallel>w\<parallel>*\<parallel>v\<parallel> + \<parallel>w\<parallel>*\<parallel>w\<parallel>" by (auto simp add: real_sq)
   also have "\<dots> = (\<parallel>v\<parallel> + \<parallel>w\<parallel>)*(\<parallel>v\<parallel> + \<parallel>w\<parallel>)" by (auto simp add: real_add_mult_distrib_ex)
   also have "\<dots> = (\<parallel>v\<parallel> + \<parallel>w\<parallel>)^2" by (auto simp add: real_sq)
   finally show ?thesis by auto
 qed
 thus ?thesis apply - apply (rule real_sq_order) apply (auto simp add: norm_pos)
   apply (rule add_nonneg_nonneg)
   by (auto simp add: norm_pos)
qed

lemma  pdist_triangle:
  "pdist A C \<le> pdist A B + pdist B C"
proof -
  have "\<parallel> loc A -: loc C \<parallel>  \<le>  \<parallel> loc A -: loc B \<parallel> + \<parallel> loc B -: loc C \<parallel>"
    proof -
      have "\<parallel> loc A -: loc C \<parallel> = \<parallel> (loc A +: zerov) -: loc C \<parallel>"
        by (auto simp add: zerov_zero_plus)
      also have "\<dots> = \<parallel> (loc A +: (loc B -: loc B)) -: loc C \<parallel>"
        by (auto simp add: minus_equal_zero)
      also have "\<dots> =  \<parallel>((loc A -: loc B) +: loc B) -: loc C\<parallel>"
        by (auto simp add: v_assoc1)
      also have "\<dots> =  \<parallel>(loc A -: loc B) +: (loc B -: loc C)\<parallel>"
        by (auto simp add: v_assoc2)
      also have "\<dots> \<le>  \<parallel>loc A -: loc B\<parallel> + \<parallel>loc B -: loc C\<parallel>"
      proof -
        have "vlen (loc A -: loc B) = sdim" by (auto simp add: minusv_def vlen_def)
        moreover
        have "vlen (loc B -: loc C) = sdim" by (auto simp add: minusv_def vlen_def)
        ultimately show ?thesis by (simp add: norm_triangle)
      qed
      finally show ?thesis by auto
    qed
  thus ?thesis by (auto simp add: pdist_def)
qed

text {* cdistl is also a pseudometric *}
lemma cdistl_noneg:
  "cdistl A B \<ge> 0"
apply (auto simp add: cdistl_def)
by (auto simp add: Ring_and_Field.mult_imp_le_div_pos vc_pos norm_pos pdist_noneg)

lemma cdistl_symm:
  "cdistl A B = cdistl B A"
by (auto simp add: vc_pos norm_pos pdist_symm cdistl_def)

lemma cdistl_triangle:
  "cdistl A C \<le> cdistl A B + cdistl B C"
proof -
  have "1/vc * pdist A C \<le>  1/vc * (pdist A B + pdist B C)"
   apply -
   apply (rule_tac c=vc in mult_left_le_imp_le)
   by (auto simp add: vc_pos pdist_triangle pdist_noneg)
  hence "1/vc * pdist A C \<le>  1/vc * pdist A B + 1/vc * pdist B C"
    by (simp only: real_add_mult_distrib2)
  thus ?thesis by (simp add: cdistl_def)
qed

text {* lower bound on direct communication distance of two agents, None if
        they can not communicate directly *}
consts
  cdistM :: "[transmitter, receiver] \<Rightarrow> real option"


definition
  cdist :: "[transmitter,receiver] \<Rightarrow> real"
 where
 "cdist T R  \<equiv> the (cdistM T R)"

text {* communication faster-than-light not possible *}
specification (cdistM)
  noflt: "cdistM (Tx A i) (Rx B j) = None \<or>
    the (cdistM (Tx A i) (Rx B j)) \<ge> cdistl A B"
  cdistnoneg: "cdistM TA RB = None  \<or> (the (cdistM TA RB) \<ge> 0)"
  by (rule_tac x="\<lambda>A B. None" in exI, auto)

lemma cdistnoneg_some:
  assumes some: "cdistM TA RB = Some y"
  shows "0 \<le> y" using some
proof -
  have "cdistM TA RB = None  \<or> (the (cdistM TA RB) \<ge> 0)" apply - by (rule cdistnoneg)
  with some have "the (cdistM TA RB) \<ge> 0" by auto
  with some show ?thesis by auto
qed

lemma noflt_some:
  assumes some: "cdistM (Tx A i) (Rx B j) \<noteq> None"
  shows         "cdistl A B \<le> the (cdistM (Tx A i) (Rx B j))"
proof -
  from noflt have or: "cdistM (Tx A i) (Rx B j) = None \<or>
                       cdistl A B \<le> the (cdistM (Tx A i) (Rx B j))" by auto
  show ?thesis
  proof (cases)
    assume "cdistM (Tx A i) (Rx B j) = None" with some show "?thesis" by contradiction
  next
    assume "cdistM (Tx A i) (Rx B j) \<noteq> None" with or show ?thesis by auto
  qed
qed

lemma noflt_some2:
  "cdistM (Tx A i) (Rx B j) = Some y \<Longrightarrow>
   cdistl A B \<le> the (cdistM (Tx A i) (Rx B j))"
  apply (insert noflt_some)
  apply auto
  apply (subgoal_tac "cdistM (Tx A i) (Rx B j) \<noteq> None") prefer 2
  apply force
  apply (drule noflt_some)
  apply auto
done

end