(*****************************************************************************

 Modeling and Verifying Properties of Security Protocols in Wireless Networks

 Module: distance/DelayedKeyDisclosure.thy (Isabelle/HOL 2009)

 Copyright (C) 2007-2009 Benedikt Schmidt <benedikt.schmidt@inf.ethz.ch>
               2008 Patrick Schaller <patrick.schaller@inf.ethz.ch>
 License: BSD (see COPYRIGHT)

******************************************************************************)

header {* Properties of hash Chains *}

theory MessageDerivationHashChains imports MessageDerivation MessageTheoryFree begin

fun (in MESSAGE_DERIVATION)
 hashn :: "nat \<Rightarrow> 'msg \<Rightarrow> 'msg" where
 "hashn 0       m = m" |
 "hashn (Suc n) m = Hash (hashn n m)"

subsection {* INITSTATE\_NOHASHCHAIN *}

locale INITSTATE_NOHASHCHAIN = INITSTATE_NONONCE_PARTS _ _ _ _ _ _ _ _ _ Key
  for Key :: "nat \<Rightarrow> 'msg" +
  fixes tag_secret :: nat and
        creator_secret :: agent and
        cn :: nat (* secret, creator, and length of hash chain *)
  assumes hashn_smaller_notin_initState:
    "!!k A. k < cn \<Longrightarrow> hashn k (Nonce creator_secret tag_secret) \<notin> parts (initState A)"
  and hash_hashn_notin_initState:
    "!!k A X. Hash (MPair (hashn k (Nonce creator_secret tag_secret)) X)
              \<notin> parts (initState A)"

begin

definition
  hashSecret :: 'msg where
  "hashSecret = Nonce creator_secret tag_secret"

lemma hashn_smaller_notin_initState2 [intro]:
  "k < cn \<Longrightarrow> hashn k hashSecret \<notin> parts (initState A)"
  apply (unfold hashSecret_def)
by (erule hashn_smaller_notin_initState)

definition
 HKey :: "nat \<Rightarrow> 'msg" where
 "HKey k =  hashn (cn - k) hashSecret"

definition
 HMAC :: "nat \<Rightarrow> 'msg \<Rightarrow> 'msg" where
 "HMAC k m = Hash[HKey k](m)"

text {* HKey k = HKey cn for k > cn because of the overflow behavior of - on nat *}

lemma HKey_overflow:
 shows "\<exists> k'. (k' - 2) \<le>  cn \<and> HKey (k - 2) = HKey (k' - 2) \<and> (k' \<le> k)"
proof cases
   assume "k - 2 \<le> cn"
   thus ?thesis apply auto done
next
  assume "\<not> (k - 2 \<le> cn)"
  hence "k - 2 > cn" by auto
  thus ?thesis
    apply (rule_tac x="cn + 2" in exI)
    apply auto
    apply (auto simp add:  HKey_def)
  done
qed

end

subsection {* Facts that hold for the free message theory *}

lemma hashn_derivable_plus:
  assumes intknold:"hashn n hashSecret \<in> DM A (knowsI A tr)"
          and "k \<ge>  0"
  shows " hashn (k+n) hashSecret \<in> DM A (knowsI A tr)"
  using intknold
  apply (induct k)
  apply simp
  apply simp
  apply (unfold DM_analz_synth)
  apply (erule synth.Hash)
done

lemma hashn_secret_minus:
  assumes nihc:"hashn n hashSecret \<notin> DM A (knowsI A tr)"
         and ksm:"n = (l::nat) + k "
  shows "hashn k hashSecret \<notin> DM A (knowsI A tr)"
  using nihc ksm
  apply -
  apply (rule notI)
  apply (frule hashn_derivable_plus)
by auto

lemma hashn_secret_smaller:
  assumes nihc:"hashn n hashSecret \<notin> DM A (knowsI A tr)"
          and ksm:"k \<le> n"
  shows "hashn k hashSecret \<notin> DM A (knowsI A tr)"
  using nihc ksm
  apply -
  apply (rule_tac l="n -k" in hashn_secret_minus)
by auto

lemma mac_synth:
  assumes msynth: "Hash \<lbrace>K,X\<rbrace> \<in> synth H"
  shows "Hash \<lbrace>K,X\<rbrace> \<in> parts H \<or> K \<in> parts (synth H)" using msynth
proof -
  from msynth have cs: "Hash \<lbrace>K,X\<rbrace> \<in> H \<or> \<lbrace>K,X\<rbrace> \<in> H \<or> K \<in> synth H"
    by auto
  thus ?thesis proof cases
    assume "Hash \<lbrace>K,X\<rbrace> \<in> H"
    hence "Hash \<lbrace>K,X\<rbrace> \<in> parts H"
      by (auto intro: parts.Fst)
   thus ?thesis by auto
 next
   assume "\<not> (Hash \<lbrace>K,X\<rbrace> \<in> H)"
   with cs have "\<lbrace>K,X\<rbrace> \<in> H \<or> K \<in> synth H" by auto
   hence "K \<in> parts (synth H)"
     by (auto intro: parts.Fst)
   thus ?thesis by auto
 qed
qed

lemma hashn_synth [rule_format]:
  assumes a: "hashn n X \<in> synth H"
  shows      "X \<in> synth H \<or> (\<exists> k. k <= n \<and> hashn k X \<in> H)" using a
proof (induct n arbitrary: X)
  case 0 show ?case using prems apply auto done
next
  case (Suc m X) show ?case using prems apply auto
    apply (drule prems(2)) apply auto
  done
qed

lemma hashn_synth_parts [rule_format]:
  assumes a: "hashn n X \<in> parts (synth H)"
  shows      "X \<in> parts (synth H) \<or> (\<exists> k. k <= n \<and> hashn k X \<in> parts H)" using a
proof -
  from a have b: "hashn n X \<in> parts H \<union> synth H"
    by auto
  show ?thesis proof cases
    assume "hashn n X \<in> parts H"
    thus ?thesis
      apply -
      apply (rule disjI2)
      apply (rule_tac x=n in exI)
      apply auto
      done
  next
    assume "hashn n X \<notin> parts H"
    with b have "hashn n X \<in> synth H" by auto
    thus ?thesis apply - apply (drule hashn_synth) by auto
  qed
qed

lemma hash_neq [simp,intro]: "Hash X \<noteq> X"
  apply (rule msg.induct, auto)
done

lemmas hash_neq2 = hash_neq [THEN not_sym]

lemma hashn_not_inf: "\<exists> n Y. X = hashn n Y \<and> Y \<notin> range Hash"
  apply (rule msg.induct) prefer 6
  apply auto
  apply (rule_tac x="n+1" in exI, auto)
  apply (rule_tac x=0 in exI, auto)+
done

lemma hash_in_out: "Hash (hashn n X) = hashn n (Hash X)"
by (induct n,auto)

lemma hash_comp [rule_format]: "\<forall>X. hashn m (hashn n X) = hashn (m + n) X"
  apply (induct n)
by (induct n, auto simp add: hash_in_out)

lemma hashn_X_neq_X: "hashn (Suc n) X \<noteq> X"
proof cases
  assume "(\<exists> n X. hashn (Suc n) X = X)"
  then obtain n X where heq: "hashn (Suc n) X = X" by auto
  from hashn_not_inf obtain m Y where "X = hashn m Y" and ynh: "Y \<notin> range Hash"
    by (fast)
  with heq have "hashn (Suc n) (hashn m Y) = hashn m Y" apply - apply auto done
  hence "hashn (Suc n) Y=Y" apply (induct m) apply auto
    apply (simp add: hash_in_out)
    done
  thus ?thesis using ynh apply auto apply (case_tac "Y::msg" ) apply auto done
next
  assume "\<not> (\<exists> n X. hashn (Suc n) X = X)"
  thus ?thesis by auto
qed

lemma range_nonce_parts: (* MOVE (MESSAGE_THEORY?)*)
  "parts (range (Nonce A)) = range (Nonce A)"
proof -
  have a: "!!x. x \<in> range (Nonce A) --> x \<in> parts (range (Nonce A))"
    by auto
  have "!!x. x \<in> parts (range (Nonce A)) ==> x \<in> range (Nonce A)"
    apply (drule parts.singleton)
    by auto
  with a show ?thesis by auto
qed

lemma range_nonce_subterms: (* MOVE (MESSAGE_THEORY?)*)
  "subterms (range (Nonce A)) = range (Nonce A)"
proof -
  have a: "!!x. x \<in> range (Nonce A) --> x \<in> subterms (range (Nonce A))"
    by auto
  have "!!x. x \<in> subterms (range (Nonce A)) ==> x \<in> range (Nonce A)"
    apply (drule subterms.singleton)
    by auto
  with a show ?thesis by auto
qed

lemma parts_analz_into_parts:
  "x \<in> parts (analz H) ==> x \<in> parts H"
by auto

lemma subterms_analz_into_subterms:
  "x \<in> subterms (analz H) ==> x \<in> subterms H"
by auto

lemma parts_hashn [rule_format, simp]:
  "parts {hashn n (Nonce A m)} = {hashn n (Nonce A m)}"
by (induct n, auto)

lemma hash_pair_hashn [simp, intro]: (* move to messHashChains *)
  "Hash \<lbrace>X,Y\<rbrace> \<noteq> hashn k (Nonce A m)"
  apply (case_tac k, auto)
  apply (case_tac nat, auto)
done

lemma hashn_inj2:
   assumes a: "k \<noteq> n"
   shows      "hashn k X \<noteq> hashn n X" using a
proof cases
  assume cp: "hashn k X = hashn n X"
  show ?thesis proof cases
    assume ge: "k > n"
    hence "\<exists> l. (Suc l) + n = k" by arith
    then obtain l where ld: "(Suc l) + n = k" apply auto done
    with prems have "hashn (Suc l) (hashn n X) = hashn k X"
      by (simp add: hash_comp)
    with prems cp have "hashn (Suc l) (hashn n X) = hashn n X" by auto
    hence "False" apply - apply (erule make_pos_goal)
      apply (rule hashn_X_neq_X) done
    thus ?thesis apply auto done
  next
    assume "\<not> (k > n)"
    with a have ge: "n > k" by arith
    hence "\<exists> l. (Suc l) + k = n" by arith
    then obtain l where ld: "(Suc l) + k = n" apply auto done
    with prems have "hashn (Suc l) (hashn k X) = hashn n X" apply -
      apply (insert prems(2) [THEN sym])
      by (simp add: hash_comp)
    with prems cp have "hashn (Suc l) (hashn k X) = hashn k X" by auto
    hence "False" apply - apply (erule make_pos_goal)
      apply (rule hashn_X_neq_X) done
    thus ?thesis apply auto done
  qed
next
  assume cp: "hashn k X \<noteq> hashn n X"
  thus ?thesis by auto
qed

lemma hashn_inj1:
   assumes a: "hashn k X = hashn n X"
   shows      "k = n" using a
  apply (rule contrapos_pp)
by (rule hashn_inj2)

subsection {* INITSTATE\_NOHASHCHAIN\_FREE *}

locale INITSTATE_NOHASHCHAIN_FREE =
  fixes initState :: "agent => msg set"
  and tag_secret :: nat and
      creator_secret :: agent and
      cn :: nat (* secret and length of hash chain *)
  assumes fhashn_smaller_notin_initState:
    "!!k A. k < cn \<Longrightarrow> hashn k (Nonce creator_secret tag_secret) \<notin> parts (initState A)"
  and fhash_hashn_notin_initState:
    "!!k A X. Hash (MPair (hashn k (Nonce creator_secret tag_secret)) X)
              \<notin> parts (initState A)"
  and fno_nonce_initState_parts [simp]: "Nonce B NB \<notin> parts (initState A)"


sublocale INITSTATE_NOHASHCHAIN_FREE
          \<subseteq> INITSTATE_NOHASHCHAIN Crypt Nonce MPair Hash Number parts subterms DM _ Key
apply (unfold_locales)
apply (rule fno_nonce_initState_parts)
apply (insert fhashn_smaller_notin_initState, auto)
apply (insert fhash_hashn_notin_initState, auto)
done

context INITSTATE_NOHASHCHAIN_FREE begin

lemma hashn_derivable:
  "hashn k hashSecret \<in> DM creator_secret H"
  apply (induct k)
  apply (simp add: hashSecret_def DM_analz_synth)
  apply (simp add: hashSecret_def DM_analz_synth)
  apply (erule synth.Hash)
done

lemma mac_derivable:
  "\<lbrace>Hash[HKey i] (Number n), HKey (i - 2)\<rbrace> \<in> DM creator_secret H"
  apply (unfold DM_analz_synth)
  apply (rule synth.MPair) prefer 2
  apply (unfold HKey_def)
  apply (rule hashn_derivable [simplified DM_analz_synth])
  apply (unfold MACM_def)
  apply (rule synth.MPair) prefer 2
  apply force
  apply (rule synth.Hash)
  apply (rule synth.MPair)
  apply (rule hashn_derivable [simplified DM_analz_synth])
  apply force
done

lemma mac_not_hashn [simp,intro]:
  "Hash[K] Y \<noteq> hashn k hashSecret"
  apply (auto simp add: hashSecret_def split: nat.split_asm)
  apply (case_tac "k")
  apply force
  apply (case_tac "nat")
  apply auto
done

lemma hashn_knowsI_hashn_received:
  assumes A: "X \<in> knowsI A tr" and 
          B: "hashn k hashSecret \<in> parts {X}" and le: "k < cn"
  shows   "\<exists> t i. (t, Recv (Rx A i) X) \<in> set tr"
  using A B le
proof -
  from A have C:"(\<exists> t i. (t, Recv (Rx A i) X) \<in> set tr) \<or> X \<in> initState A"
    apply auto
    apply (drule knowsI_A_imp_Recv_initState)
    apply fast
    done
  have InHa:"!!k n X. k \<noteq> n \<Longrightarrow> hashn k X \<noteq> hashn n X"  by (rule hashn_inj2)
  from B le InHa have D:"hashn k hashSecret \<notin> parts (initState A)" apply -
    apply (rule hashn_smaller_notin_initState2)
    by auto
  from B D have E:"X \<notin> initState A"
    apply -
    apply (rule notI)
    apply (rotate_tac)
    apply (erule make_neg_goal)
    apply (auto intro: parts.elem_trans)
    done
  from A B C D E show ?thesis
    by auto
qed

lemma  hashn_derivI_hashn_received:
  assumes deriv: "hashn n hashSecret \<in> parts (DM A (knowsI A tr))" and le: "n < cn" and
          notBr: "A \<noteq> creator_secret"
  shows   "\<exists> X t i j. (t, Recv (Rx A i) X) \<in> set tr \<and>
                      hashn j hashSecret \<in> parts {X} \<and> j <= n"
  using deriv le
proof -
  let ?H = "range (Nonce A) \<union> analz (knowsI A tr)"
  from deriv have o: "hashSecret \<in> parts (synth ?H) \<or> (\<exists>k\<le>n. hashn k hashSecret \<in> parts ?H)"
    apply -
    apply (rule hashn_synth_parts)
    apply (unfold DM_analz_synth)
    apply auto
    done
  show ?thesis proof cases
    assume "hashSecret \<in> parts (synth ?H)"
    hence hun: "hashSecret \<in> parts ?H \<union> synth ?H" by auto
    show ?thesis proof cases
      assume "hashSecret \<in> synth ?H"
      hence p: "Nonce creator_secret tag_secret \<in> DM A (knowsI A tr)"
       apply (auto simp add: hashSecret_def DM_analz_synth) done
      have q: "Nonce creator_secret tag_secret \<in> parts {Nonce creator_secret tag_secret}"
	by auto
      with p notBr show ?thesis
        apply -
        apply (drule othernonce_gen_received_parts)
        apply auto
        apply (rule_tac x=Y in exI)
        apply auto
        apply (rule_tac x=0 in exI)
        apply (auto simp add: hashSecret_def)
	done
    next
      assume "hashSecret \<notin> synth ?H"
      with hun have "hashSecret \<in> parts ?H" by auto
      hence p: "Nonce creator_secret tag_secret \<in> parts (analz (knowsI A tr))"
	using notBr
	apply (auto simp add: hashSecret_def DM_analz_synth)
	apply (drule parts.singleton) apply (auto)
	done
      hence "(Nonce creator_secret tag_secret) \<in> parts (knowsI A tr)" apply -
        by (rule parts_analz_into_parts)
      hence ex: "\<exists> X \<in> knowsI A tr. (Nonce creator_secret tag_secret) \<in> parts {X}"
        apply (rule parts.singleton)
        done
      have q: "Nonce creator_secret tag_secret \<in> parts {Nonce creator_secret tag_secret}"
	by auto
      from ex obtain X where "X \<in> knowsI A tr"
	                 and "(Nonce creator_secret tag_secret) \<in> parts {X}" by auto
      with p notBr show ?thesis
        apply -
        apply (drule othernonce_gen_received_parts)
        apply force
        apply (force simp add: DM_analz_synth)
        apply (erule exE)+
        apply (rule_tac x=Y in exI)
        apply (rule_tac x=t in exI)
        apply (rule_tac x=i in exI)
        apply (rule_tac x=0 in exI)
        apply (auto simp add: hashSecret_def)
      done
    qed
  next
    assume "hashSecret \<notin> parts (synth ?H)"
    with o have "\<exists>k\<le>n. hashn k hashSecret \<in> parts (?H)" by auto
    then obtain k where "hashn k hashSecret \<in> parts (?H)" and kleq: "k <= n" by auto
    with notBr have "hashn k hashSecret \<in> parts (analz (knowsI A tr))" apply -
      apply (simp add: range_nonce_subterms)
      apply (auto split: msg.split nat.split) apply (case_tac "k::nat")
      apply (auto simp add: hashSecret_def)
      apply (drule parts.singleton) apply (auto) 
      apply (drule parts.singleton) apply (auto)
      done
    hence  "hashn k hashSecret \<in> parts (knowsI A tr)" apply -
      by (rule parts_analz_into_parts)
    hence "\<exists> X. X \<in> knowsI A tr \<and> hashn k hashSecret \<in> parts {X}" apply -
      apply (drule parts.singleton) by auto
    then obtain X where p: "X \<in> knowsI A tr \<and> hashn k hashSecret \<in> parts {X}" by auto
    from le kleq have kle: "k < cn" by arith
    with p have "\<exists>t i. (t, Recv (Rx A i) X) \<in> set tr" apply -
      apply (rule hashn_knowsI_hashn_received) apply auto done
    with p kleq show ?thesis apply auto done
  qed
qed

lemma mac_derivI_mac_or_hashn_received:
  assumes deriv: "Hash\<lbrace>hashn n hashSecret,X\<rbrace> \<in> parts (DM A (knowsI A tr))" and
          le: "n < cn" and notBr: "A \<noteq> creator_secret"
  shows   "(\<exists> Y t i j. (t, Recv (Rx A i) Y) \<in> set tr \<and>
                       hashn j hashSecret \<in> parts {Y} \<and> j <= n) \<or>
           (\<exists> Y t i . (t, Recv (Rx A i) Y) \<in> set tr \<and>
                      Hash \<lbrace>hashn n hashSecret,X\<rbrace> \<in> parts {Y})"
  using deriv le
proof -
  let ?H = "range (Nonce A) \<union> analz (knowsI A tr)" and
      ?MAC = "Hash \<lbrace>hashn n hashSecret, X\<rbrace>"
   
  from deriv have o: "?MAC \<in> parts ?H \<or> ?MAC \<in> synth ?H"
    apply (unfold DM_analz_synth)
    by (simp)
  show ?thesis proof cases
    assume "?MAC \<in> parts ?H"
    hence "?MAC \<in> parts (knowsI A tr)" by (auto dest: parts.singleton)
    thus ?thesis using le apply -
      apply (rule disjI2)
      apply (drule parts_knowsI)
      apply auto
      apply (insert hash_hashn_notin_initState)
      apply (auto simp add: hashSecret_def)
      done
  next
    assume "?MAC \<notin> parts ?H"
    with o have "?MAC \<in> synth ?H" by auto
    hence p: "Hash \<lbrace>hashn n hashSecret,X\<rbrace> \<in> parts ?H \<or> hashn n hashSecret \<in> parts (synth ?H)"
      by (intro mac_synth)
    show ?thesis proof cases
      assume "Hash \<lbrace>hashn n hashSecret,X\<rbrace> \<in> parts ?H"
      hence "Hash \<lbrace>hashn n hashSecret,X\<rbrace> \<in> parts (knowsI A tr)"
	by (auto dest: parts.singleton)
      thus ?thesis apply -
        apply (rule disjI2)
        apply (drule parts_knowsI)
	apply (insert hash_hashn_notin_initState)
	apply (auto simp add: hashSecret_def)
        done
    next
      assume "Hash \<lbrace>hashn n hashSecret,X\<rbrace> \<notin> parts ?H"
      with p have " hashn n hashSecret \<in> parts (synth ?H)" by auto
      hence "hashn n hashSecret \<in> parts (DM A (knowsI A tr))"
        apply (auto simp add: DM_analz_synth)
        done
       thus ?thesis using le notBr apply -
         apply (drule hashn_derivI_hashn_received, auto)
         done
     qed
   qed
qed

end

end
