(*  Title:      HOL/Auth/Public
    ID:         $Id: Public.thy,v 1.35 2007/08/01 18:25:16 wenzelm Exp $
    Author:     Lawrence C Paulson, Cambridge University Computer Laboratory
                Benedikt Schmidt, ETH Zurich
    Copyright   1996  University of Cambridge
                2008-9 ETH Zurich

Theory of Public Keys (common to all public-key protocols)

Private and public keys; 
*)

header {* Initial knowledge of Agents (Key distributions) *}

theory Public imports Event MessageTheory NatEmbed begin

subsection{*Asymmetric Keys*}

datatype keymode = Signature | Encryption

consts
  publicKey :: "[keymode,agent] => key"

abbreviation
  pubEK :: "agent => key" where
  "pubEK == publicKey Encryption"

abbreviation
  pubSK :: "agent => key" where
  "pubSK == publicKey Signature"

abbreviation
  privateKey :: "[keymode, agent] => key" where
  "privateKey b A == invKey (publicKey b A)"

abbreviation
  (*BEWARE!! priEK, priSK DON'T WORK with inj, range, image, etc.*)
  priEK :: "agent => key" where
  "priEK A == privateKey Encryption A"

abbreviation
  priSK :: "agent => key" where
  "priSK A == privateKey Signature A"

text{* The function symKey returns for every pair agents a shared
secret key. The axiom symmetric-SymKey ensures that the returned key
is a symmetric key.*}

consts symKey :: "[agent,agent] \<Rightarrow> key"

axioms
 --{*The keys returned by the function symKey are symmetric keys*}
 symmteric_SymKey[simp]: "invKey (symKey A B) = symKey A B"

specification(symKey)
  injective_symKey:
    "symKey A B = symKey C D \<Longrightarrow> {A,B} = {C,D}"
  com_SymKey:
  "{A,B} = {C,D} \<Longrightarrow> symKey A B = symKey C D"
 apply (rule exI [of _ "%A B. prod (primefactors (agent_case (\<lambda> n. 2*n) (\<lambda> m. 2*m +1) A) (agent_case (\<lambda> n. 2*n) (\<lambda> m. 2*m +1) B))"])
 apply (rule conjI)
 apply (rule allI)+
 apply (rule impI)
 apply (erule  f_inj_prod_inj)
 apply simp
 apply (simp add: inj_on_def split: agent.split)
 apply auto
 apply arith+
 apply (erule f_inj_primef_eq )
 apply (simp add: inj_on_def split: agent.split)
 apply auto
 apply (arith)+
done


text{*By freeness of agents, no two agents have the same key.  Since
  @{term "True\<noteq>False"}, no agent has identical signing and encryption keys*}
specification (publicKey)
  injective_publicKey:
    "publicKey b A = publicKey c A' ==> b=c & A=A'"
   apply (rule exI [of _ 
       "%b A. agent_case (\<lambda>n. n*4) (\<lambda>n. n*4 +2)  A + keymode_case 0 1 b"])
   apply (auto simp add: inj_on_def split: agent.split keymode.split)
   apply arith +
   done

axioms
  (*No private key equals any public key (essential to ensure that private
    keys are private!) *)
  privateKey_neq_publicKey [iff]: "privateKey b A \<noteq> publicKey c A'"
  privateKey_neq_symKey [iff]: "privateKey b A \<noteq> symKey C D"
  pubKey_neq_symKey [iff]: "publicKey b A \<noteq> symKey C D"

lemmas publicKey_neq_privateKey = privateKey_neq_publicKey [THEN not_sym]
declare publicKey_neq_privateKey [iff]

lemmas symKey_neq_privateKey = privateKey_neq_symKey [THEN not_sym]
declare symKey_neq_privateKey [iff]

lemmas symKey_neq_publicKey = privateKey_neq_symKey [THEN not_sym]
declare symKey_neq_publicKey [iff]

lemma publicKey_inject [iff]: "(publicKey b A = publicKey c A') = (b=c & A=A')"
by (blast dest!: injective_publicKey)

subsubsection{*Inverse of keys *}

lemma invKey_eq [simp]: "(invKey K = invKey K') = (K=K')"
  apply safe
  apply (drule_tac arg_cong [where f=invKey], simp)
done

lemma invKey_image_eq [simp]: "(invKey x \<in> invKey`A) = (x \<in> A)"
  apply auto
done

(*holds because invKey is injective*)
lemma publicKey_image_eq [simp]:
  "(publicKey b x \<in> publicKey c ` AA) = (b=c & x \<in> AA)"
by auto

lemma privateKey_notin_image_publicKey [simp]: "privateKey b x \<notin> publicKey c ` AA"
by auto

lemma privateKey_image_eq [simp]:
  "(privateKey b A \<in> invKey ` publicKey c ` AS) = (b=c & A\<in>AS)"
by auto

lemma publicKey_notin_image_privateKey [simp]:
  "publicKey b A \<notin> invKey ` publicKey c ` AS"
by auto 

subsection{* Locales for Public Key Distribution, Shared Symmetric Keys, and Nonces *}
locale INITSTATE_PKSIG = INITSTATE _ _ _ _ _ _ _ _ Key for Key :: "nat \<Rightarrow> 'msg" +
  assumes priSK_known_self: "Key (priSK A) \<in> initState A"
  assumes priSK_notknown_other_subterms: "A \<noteq> B \<Longrightarrow> Key (priSK B) \<notin> subterms (initState A)"
  assumes pubSK_known: "Key (pubSK A) \<in> initState B"
  assumes priSK_not_used: "Crypt (priSK A) X \<notin> subterms (initState B)"
  (* XXX: perhaps change subterms to also return Key k for Crypt k m, would merge 2. and 4. assm *)

lemma (in INITSTATE_PKSIG) priSK_notknown_other:
  "A \<noteq> B \<Longrightarrow> Key (priSK B) \<notin> initState A"
  apply auto
  apply (subgoal_tac "Key (priSK B) \<notin> subterms (initState A)")
  prefer 2
  apply (rule priSK_notknown_other_subterms)
  apply force
  apply (rotate_tac 2)
  apply (erule contrapos_np)
  apply (erule subsetD2)
  apply (rule subterms.increasing)
done

locale INITSTATE_PKENC = INITSTATE  _ _ _ _ _ _ _ _ Key for Key :: "nat \<Rightarrow> 'msg" +
  assumes priEK_known_self: "Key (priEK A) \<in> initState A"
  assumes priEK_notknown_other_subterms: "A \<noteq> B \<Longrightarrow> Key (priEK B) \<notin> subterms (initState A)"
  assumes pubEK_known: "Key (pubEK A) \<in> initState B"
  assumes priEK_not_used: "Crypt (priEK A) X \<notin> subterms (initState B)"


lemma (in INITSTATE_PKENC) priEK_notknown_other:
  "A \<noteq> B \<Longrightarrow> Key (priEK B) \<notin> initState A"
  apply auto
  apply (subgoal_tac "Key (priEK B) \<notin> subterms (initState A)")
  prefer 2
  apply (rule priEK_notknown_other_subterms)
  apply force
  apply (rotate_tac 2)
  apply (erule contrapos_np)
  apply (erule subsetD2)
  apply (rule subterms.increasing)
done

locale INITSTATE_SYMKEYS = INITSTATE  _ _ _ _ _ _ _ _ Key for Key :: "nat \<Rightarrow> 'msg" +
  assumes symKey_known_self: "!!B. Key (symKey A B) \<in> initState A"
  assumes symKey_notknown_other_subterms:
    "\<lbrakk> A \<noteq> B; A \<noteq> C \<rbrakk> \<Longrightarrow> Key (symKey B C) \<notin> subterms (initState A)"
  assumes symKey_not_used: "Crypt (symKey A B) X \<notin> subterms (initState C)"
  assumes symKey_not_used_MAC: "Hash (MPair (Key (symKey A B)) X) \<notin> subterms (initState C)"

lemma (in INITSTATE_SYMKEYS) priEK_notknown_other:
  "\<lbrakk> A \<noteq> B; A \<noteq> C \<rbrakk> \<Longrightarrow> Key (symKey B C) \<notin> initState A"
  apply auto
  apply (subgoal_tac "Key (symKey B C) \<notin> subterms (initState A)")
  prefer 2
  apply (erule symKey_notknown_other_subterms)
  apply force
  apply (rotate_tac 2)
  apply (erule contrapos_np)
  apply (erule subsetD2)
  apply (rule subterms.increasing)
done

locale INITSTATE_NONONCE = INITSTATE _ _ _ _ _ _ _ _ Key for Key :: "nat \<Rightarrow> 'msg" +
  assumes no_nonce_initState_subterms [simp]: "Nonce B NA \<notin> subterms (initState A)"

lemma (in INITSTATE_NONONCE) no_nonce_initState:
  "Nonce B NA \<notin> initState A"
  apply auto
  apply (subgoal_tac "Nonce B NA \<notin> subterms (initState A)")
  prefer 2
  apply (rule no_nonce_initState_subterms)
  apply (rotate_tac 2)
  apply (erule contrapos_np)
  apply (erule subsetD2)
  apply (rule subterms.increasing)
done

lemma (in INITSTATE_NONONCE) nonce_knowsI_nonce_received:
  assumes A: "X \<in> knowsI A tr" and
          B: "Nonce B NA \<in> subterms {X}"
  shows   "\<exists> t i. (t, Recv (Rx A i) X) \<in> set tr"
  using A B
proof -
  from A have C: "(EX t i. (t, Recv (Rx A i) X) \<in> set tr) \<or> X \<in> initState A"
    by (intro knowsI_A_imp_Recv_initState, auto)
  let ?A = "EX t i. (t, Recv (Rx A i) X) \<in> set tr"
  show ?thesis
  proof cases
    assume ?A
    thus ?thesis by auto
  next
    assume "\<not> ?A"
    with C have D: "X \<in> initState A" by clarsimp
    have "Nonce B NA \<notin> initState A"
      apply auto
      apply (drule subterms.inj)
      apply (erule contrapos_pp)
      apply (rule no_nonce_initState_subterms)
      done
    with B have "X \<notin> initState A"
      apply auto
      apply (subgoal_tac "Nonce B NA \<in> subterms (initState A)")
      apply force
      apply (erule_tac G="{X}" in subterms.trans)
      apply force
      done
    with D show ?thesis by contradiction
  qed
qed

lemma (in INITSTATE) subterms_knowsI:
  "X \<in> subterms (knowsI A tr) \<Longrightarrow>
   (\<exists> t Y i. (t, Recv (Rx A i) Y) \<in> set tr \<and> X \<in> subterms {Y}) \<or> X \<in> subterms (initState A)"
  apply (drule subterms.singleton)
  apply auto
  apply (drule knowsI_A_imp_Recv_initState)
  apply auto
  apply (drule_tac subterms.inj, drule_tac H="initState A" in subterms.trans, auto)
done

lemma (in INITSTATE) parts_knowsI:
  "X \<in> parts (knowsI A tr) \<Longrightarrow>
   (\<exists> t Y i. (t, Recv (Rx A i) Y) \<in> set tr \<and> X \<in> parts {Y}) \<or> X \<in> parts (initState A)"
  apply (drule parts.singleton)
  apply auto
  apply (drule knowsI_A_imp_Recv_initState)
  apply auto
  apply (drule_tac parts.inj, drule_tac H="initState A" in parts.trans, auto)
done

locale INITSTATE_NONONCE_PARTS = INITSTATE _ _ _ _ _ _ _ _ Key for Key :: "nat \<Rightarrow> 'msg" +
  assumes no_nonce_initState_parts [simp]: "Nonce B NA \<notin> parts (initState A)"

lemma (in INITSTATE_NONONCE_PARTS) no_nonce_initState:
  "Nonce B NA \<notin> initState A"
  apply auto
  apply (subgoal_tac "Nonce B NA \<notin> parts (initState A)")
  prefer 2
  apply (rule no_nonce_initState_parts)
  apply (rotate_tac 2)
  apply (erule contrapos_np)
  apply (erule subsetD2)
  apply (rule parts.increasing)
done

lemma (in INITSTATE_NONONCE_PARTS) nonce_knowsI_nonce_received_parts:
  assumes A: "X \<in> knowsI A tr" and
          B: "Nonce B NA \<in> parts {X}"
  shows   "\<exists> t i. (t, Recv (Rx A i) X) \<in> set tr"
  using A B
proof -
  from A have C: "(EX t i. (t, Recv (Rx A i) X) \<in> set tr) \<or> X \<in> initState A"
    by (intro knowsI_A_imp_Recv_initState, auto)
  let ?A = "EX t i. (t, Recv (Rx A i) X) \<in> set tr"
  show ?thesis
  proof cases
    assume ?A
    thus ?thesis by auto
  next
    assume "\<not> ?A"
    with C have D: "X \<in> initState A" by clarsimp
    have "Nonce B NA \<notin> initState A"
      apply auto
      apply (drule parts.inj)
      apply (erule contrapos_pp)
      apply (rule no_nonce_initState_parts)
      done
    with B have "X \<notin> initState A"
      apply auto
      apply (subgoal_tac "Nonce B NA \<in> parts (initState A)")
      apply force
      apply (erule_tac G="{X}" in parts.trans)
      apply force
      done
    with D show ?thesis by contradiction
  qed
qed

end